******************************************************************************
*
*       Function:   GET_PROC
*       Usage:      proc_idx = get_proc (stream, last_status)
*       Purpose:    This routine obtains a subprocess for exclusive use
*                   by 'stream'.  It first looks for an available, already
*                   running process.  If none are available, then a new
*                   process is created.
*       Output:     The return value is the index into procs() of the process
*                   allocated.  If the is -1 then no process was allocated,
*                   and 'lstat' contains the last status value that caused
*                   the process to not be available.
*
******************************************************************************

        integer function get_proc (stream, lstat)
            implicit none

            include 'defs.inc'
            include 'struct.inc'
            include 'extern.inc'
            include 'global.inc'

            integer stream, i, lstat, avail, next_avail, stat
            integer sys$waitfr
            external sys$waitfr

*           Initially, no process found

            get_proc = -1
            avail = -1
            next_avail = -1

            lstat = SS$_NOMOREPROC

*           Search through all processes slots.  Save the index of the first
*           non-started process, and always prefer to use a process that is
*           started, even if there is a slot preceeding it that is available,
*           but does not have a process started.  This makes sure that we
*           always use started processes, before starting new ones.
                           
            do i=1,MAXPROCS

*               If the slot is NOT in use by another stream then check it.

                if (procs(i).avail) then

*                   If a process is running and it is ours, then use it first.

                    if (procs(i).started .and. procs(i).dedicated .and.
     .                      (procs(i).stream .eq. stream) .and.
     .                      (procs(i).uic .eq. streams(stream).uic)) then

                        get_proc = i                           
                        procs(i).avail = .false.
                        return

*                   If there is no process running, then save this index,
*                   so that we may use the slot if no other processes are
*                   running that are available.

                    else if (.not. procs(i).started) then
                        if (avail .eq. -1) avail = i
*
*                   Otherwise if this is a generic queue, and a generic,
*                   running process, then use it.
*
                    else if ((procs(i).uic .eq. streams(stream).uic) .and.
     .                      (.not. procs(i).dedicated) .and.
     .                      (.not. streams(stream).special)) then

                        get_proc = i
                        procs(i).stream = stream
                        procs(i).avail = .false.
                        return
                    else

*                       If available and running, but does not have the right
*                       attributes, make it a candidate for deletion.

                        next_avail = i
                    endif
                endif
            end do

*           If we did not find a running process, then start one in the
*           first available slot that we found.

            if (avail .ne. -1) then
                get_proc = avail
                procs(avail).stream = stream
                procs(avail).avail = .false.
            else if (next_avail .ne. -1) then

*               Otherwise, snuff an idle process, and use it's slot
*               When this occurs, it generally indicates that either
*               the symbiont should be recompiled with more process
*               slots, or the timeout value should be decreased on
*               queues using the user=user configuration parameter,
*               or process slots are not being freed properly.

*               Get an event flag for the AST procedure to set.

                call signal (lib$get_ef (procs(next_avail).alloc_ef))

*               Delete the idle process.

                procs(next_avail).dying = .true.
                stat = sys$delprc (procs(next_avail).pid,)
                if ((stat.and.1).ne.1) then
                    call log_str (%val (stat),
     .                   'Error deleting process when out of process slots')
                endif

*               Log a message to the console to notify people.

                call log_sys (%val (SS$_NORMAL),
     .              'Waiting for a process slot, perhaps more should be added')

*               Wait for the deletion.   

                call signal (sys$waitfr (%val(procs(next_avail).alloc_ef)))

*               Reinitialize the structure for our use.

                procs(next_avail).stream = stream
                procs(next_avail).avail = .false.

*               Free the event flag
                
                call signal (lib$free_ef (procs(next_avail).alloc_ef))

*               Invalidate the event flag number

                procs(next_avail).alloc_ef = -1
            endif
        end
